// ==================================================================
//   THEREMINO SYSTEM - Application Spectrosensors - Nano Firmware
// ==================================================================
//  Firmare for sensors Toshiba TCD1304 and TCD1254 
//  Using a Nano module with CH340 (do not use FTDI chip models)
// ==================================================================
//   Nano Pin      Sensor Pin 
//    D9            SH
//    D10           MCLK
//    D11           ICG
//    A0            OS
//    3.3V          VAD and VDD (with 10uF or 47uF and 100 nF to GND)
//    GND           SS (and NC pins from 6 to 20)
// ------------------------------------------------------------------
//  The Nano pin AREF must be connected to 3.3V
// ==================================================================

#include <util/delay.h>

#define SH   0x02       // D9
#define MCLK 0x04       // D10
#define ICG  0x08       // D11

#define CLOCK PORTB     // PORTB is D8 to D13

#define BUFFER_SIZE 272 // Normally 272 (17x16) (200 is not OK) (working from 9 to 550)

#define DUMMY_PIXELS 32 // This jumps the first sensor pixels (can be set to 32 or 24) 

uint16_t buffer[BUFFER_SIZE];

uint16_t NumSamples = 3600;
uint8_t  AdcSpeed = 3;
uint8_t  DebugType = 0;
uint8_t  SensorType = 0; // 0=TCD1304 1=TCD1254 

byte     AdcDummyCounter;
uint16_t AdcCounter;
bool     AdcRestartFlag = false;

uint16_t BufferWriteIndex;
uint16_t BufferReadIndex;

uint32_t ExposureTime = 1000;
uint32_t ExposureStartTime;

// ==================================================================
//   SETUP
// ==================================================================
void setup() 
{
  // ------------------------------------------------------ Enable the serial port.
  Serial.begin(1000000);
  // ------------------------------------------------------ Set the ADC refernce
  //analogReference(DEFAULT);  // 5V AnalogVCC
  //analogReference(INTERNAL); // 1.1V internal
  analogReference(EXTERNAL);   // External pin AREF
  delay(10); 
  // ------------------------------------------------------ INIT CLOCKS
  DDRB |= (SH | ICG | MCLK);  // clock lines to outputs
  CLOCK |= ICG;               // integration clear gate high
  // ------------------------------------------------------ INIT NumSamples and AdcSpeed
  SetNumPixelsAndAdcSpeed();
  // ------------------------------------------------------ INIT flags
  AdcRestartFlag = true;
}

void SetSensorClock(uint32_t frequency)
{
  TCCR1A = (1<<COM1B1) | (0<<COM1B0) | (1<<WGM11); // Fast PWM mode, PWM not inverted on OC1B (Pin D10)
  TCCR1B = (1<<WGM13) | (1<<WGM12) | (0<<CS12) | (0<<CS11) | (1<<CS10); // Prescaler = 1
  //
  uint32_t icr_value = F_CPU / frequency; // Computing ICR1 for the desired frequency
  ICR1 = icr_value;
  OCR1A = icr_value / 2;
  OCR1B = icr_value / 2;
  //
  DDRB |= (1<<PB2); // Set PB2 (D10) as output
}

void SetAdcSpeed(uint8_t AdcSpeed)
{
  // ------------------------------------
  //  ADPS2 ADPS1 ADPS0	Division Factor
  // ------------------------------------
  //   0	    0	    0      2
  //   0	    0     1	     2
  //   0	    1	    0	     4
  //   0	    1	    1	     8
  //   1	    0	    0	     16
  //   1	    0	    1	     32
  //   1	    1     0	     64
  //   1	    1     1	     128
  // ----------------------------------------------------- Reset prescaler bits
  ADCSRA &= ~((1<<ADPS2) | (1<<ADPS1) | (1<<ADPS0));
  // ----------------------------------------------------- Set prescaler bits
  if      (AdcSpeed == 4) ADCSRA |= (1<<ADPS2) | (0<<ADPS1) | (0<<ADPS0); // ADC prescaler = 16
  else if (AdcSpeed == 3) ADCSRA |= (1<<ADPS2) | (0<<ADPS1) | (1<<ADPS0); // ADC prescaler = 32  SPS=38461Hz  26uS/sample
  else if (AdcSpeed == 2) ADCSRA |= (1<<ADPS2) | (1<<ADPS1) | (0<<ADPS0); // ADC prescaler = 64
  else                    ADCSRA |= (1<<ADPS2) | (1<<ADPS1) | (1<<ADPS0); // ADC prescaler = 128
  // ----------------------------------------------------- Select ADC channel to A0
  ADMUX &= ~(1<<MUX3 | 1<<MUX2 | 1<<MUX1 | 1<<MUX0);
  // ----------------------------------------------------- Enable ADC and interrupts
  ADCSRA |= (1<<ADEN); // Enable ADC
  ADCSRA |= (1<<ADIE); // Enable ADC interrupts
}

// uint32_t SensorClock;
// void SetNumPixelsAndAdcSpeed()
// {
//   SensorClock = 553845500 / NumSamples;
//   if (AdcSpeed == 4) SensorClock *= 2;
//   if (AdcSpeed == 2) SensorClock /= 2;
//   if (AdcSpeed == 1) SensorClock /= 4;
//   if (SensorType == 1) SensorClock = (SensorClock * 2500) / 3600 / 2 ;  // 0=TCD1304 1=TCD1254 
//   SetSensorClock(SensorClock);
//   SetAdcSpeed(AdcSpeed);
// }

uint32_t SensorClock;
void SetNumPixelsAndAdcSpeed()
{
  if (SensorType == 1) SensorClock = 192307700 / NumSamples; // 1=TCD1254 (2500 pixels and 2 clock per pixel)
  else                 SensorClock = 553846176 / NumSamples; // 0=TCD1304 (3600 pixels and 4 clock per pixel)
  if (AdcSpeed == 4) SensorClock *= 2;
  if (AdcSpeed == 2) SensorClock /= 2;
  if (AdcSpeed == 1) SensorClock /= 4;
  SetSensorClock(SensorClock);
  SetAdcSpeed(AdcSpeed);
}

// ---------------------------------------------------------------------------------
//   Sensor     Clock Hz      Pixel      Clock/pixel     SensorClock base number
// ---------------------------------------------------------------------------------
//   TCD1254    38461.54     *  2500   *    2        =   192.307.700
//   TCD1304    38461.54     *  3600   *    4        =   553.846.176

// ---------------------------------------------------------------------------------
//   Clock Hz  Pixel Hz  Pixel uS  3600 pixels  ADC Hz    ADC uS     Samples
// ---------------------------------------------------------------------------------
//   381000     95250      10.5      38 mS     38461.54      26       1453 (800 error by Dave Allmon ?)
//   153846     38444      26      93.6 mS     38461.54      26       3615

//             --------------------- Sensor Clock -----------------------
//  nPixels    AdcSpeed-4  AdcSpeed-3  AdcSpeed-2  AdcSpeed-1  
// ---------------------------------------------------------------------------------
//   3600                   153846 Hz     76923 Hz    38461 Hz    
//   3000                   184615 Hz     92307 Hz    46154 Hz    
//   2500                   221538 Hz    110769 Hz    55384 Hz    
//   2000                   276923 Hz    138461 Hz    69231 Hz   
//   1500                   369230 Hz    184615 Hz    92307 Hz   
//   1000                   553845 Hz    276923 Hz   138461 Hz   
//    500                  1107691 Hz    553845 Hz   276923 Hz   

//               ----------- Frames/sec (approx) --------------
//  nPixels     AdcSpeed-4  AdcSpeed-3  AdcSpeed-2  AdcSpeed-1
// -------------------------------------------------------------
//   3600            20       10 (94mS)       5         2.5     
//   3000            24       12              6           3       
//   2500            30       15              8           4       
//   2000            50       25             12           6       
//   1500            68       34             17           8      
//   1000           102       51             25          12     
//    500           204      102 (9.8mS)     51          25     

// ---------------------------------------------------------------------------------
//  AdcSpeed  ADC Clock  ADC-ClockFreq   ADC-Clock  uS/sample  Samplig Frequency
// ---------------------------------------------------------------------------------
//    4       16MHz/16      1 MHz         1 uS        13 uS        76923.08 
//    3       16MHz/32    500 KHz         2 uS        26 uS        38461.54 
//    2       16MHz/64    250 KHz         4 uS        52 uS        19230.77
//    1       16MHz/128   125 KHz         8 uS       104 uS         9615.38

// ---------------------------------------------------------------------------------
// SENSOR   nPixel   Length     Pixel-size    Distance    Lens
// ---------------------------------------------------------------------------------
// TCD1304   3600    29.1 mm        8 um       7.16 mm   50-60 mm
// TCD1254   2500    13.13 mm    5.25 um       2.23 mm   15-20 mm


// ==================================================================
//   LOOP
// ==================================================================
void loop() 
{
  SendSamples();
  if (AdcRestartFlag)
  {
    AdcRestartFlag = false;
    ReadCommands();
    RestartAdc();
  }
}

// ==================================================================
//   ADC INTERRUPT
// ==================================================================
const byte StartSequence[] = {255, 254, 253, 252, 251, 250, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9};

ISR(ADC_vect) 
{
  if (AdcCounter < NumSamples)
  {
    if (AdcDummyCounter < DUMMY_PIXELS) // jump first pixels (normally 32 or 24)
    {
      AdcDummyCounter++;
    }
    else
    {
      //CLOCK |= SH;
      switch (DebugType)
      {
        case 0: buffer[BufferWriteIndex] = ADC; break; // do not use the slow analogRead() function
        case 1: buffer[BufferWriteIndex] = AdcCounter & 1023; break;
        case 2: buffer[BufferWriteIndex] = AdcCounter / 16 & 1023; break;
        case 3: buffer[BufferWriteIndex] = AdcCounter / 64 & 1023; break;
        case 4: buffer[BufferWriteIndex] = AdcCounter / 16 * AdcCounter / 64 & 1023; break;
        case 5: buffer[BufferWriteIndex] = (AdcCounter / 16 * AdcCounter) / 256 & 1023; break;
        case 6: buffer[BufferWriteIndex] = (AdcCounter / 16 * AdcCounter) / 4096 & 1023; break;
        case 7: buffer[BufferWriteIndex] = AdcCounter * AdcCounter / 64 & 1023; break;
        case 8: buffer[BufferWriteIndex] = (AdcCounter * AdcCounter) / 256 & 1023; break;
        case 9: buffer[BufferWriteIndex] = (AdcCounter * AdcCounter) / 4096 & 1023; break;
      }
      //CLOCK &= ~SH;
      AdcCounter ++;
      if (++BufferWriteIndex >= BUFFER_SIZE) BufferWriteIndex = 0;
    }
  }
  else
  {
    ADCSRA &= ~(1<<ADATE); // Disable ADC free-running
    ADCSRA &= ~(1<<ADEN);  // Disable ADC
    AdcRestartFlag = true;
  }
}

// ==================================================================
//   RESTART ADC
// ==================================================================
void RestartAdc(void) 
{
  // -------------------------------------------------- Disable Interrupts
  noInterrupts(); 

  // -------------------------------------------------- Try to stabilize clock
  // while (digitalRead(10));
  // SetSensorClock(SensorClock);
  // TCCR1B &= ~(1 << CS20); // Stop the CLOCK timer
  // // TCNT1 = 1;              // Set the CLOCK TIMER counter to 1
  // // TCCR1B |= (1 << CS20);  // Restart the CLOCK TIMER
  // // //CLOCK &= ~MCLK;
  // // //CLOCK |= SH;
  // // TCCR1B &= ~(1 << CS20); // Stop the CLOCK timer

  // -------------------------------------------------- Signal the exposure start
  CLOCK |= SH;
  delayMicroseconds(50);   // 5  ( was 5 )
  CLOCK &= ~SH;

  // -------------------------------------------------- EXPOSURE time microsec
  if (ExposureTime < 15000) // < 15 mS
  {               
    delayMicroseconds(ExposureTime - 5); // delayMicroseconds max value is 16000
  }
  else
  {
    for (uint32_t i = 0; i < ExposureTime / 5000; i++)
    {
      interrupts();
      delayMicroseconds(4900); // 4900 to compensate the ReadCommands overhead
      ReadCommands();
      noInterrupts();
    }
  }
 
  // -------------------------------------------------- Sync the CLOCK start
  TCCR1B &= ~(1 << CS20); // Stop the CLOCK timer
  TCNT1 = 1;              // Set the CLOCK TIMER counter to 1
  TCCR1B |= (1 << CS20);  // Restart the CLOCK TIMER

  // -------------------------------------------------- PULSES ICG and SH 
  // -------------------------------------------------- Send signals to CCD (the SH signals END OF EXPOSURE)
  CLOCK &= ~ICG;
  _delay_loop_1(80);      // 80 ( was 12 = 2.5 uS )
  CLOCK |= SH;
  delayMicroseconds(5);   // 5  ( was 5 )
  CLOCK &= ~SH;
  delayMicroseconds(15);  // 15 ( was 15 )
  CLOCK |= ICG;
  delayMicroseconds(1);   // 1  ( was 1 )
  // -------------------------------------------------- Add START SYNC BYTES
  for (int i = 0; i <= 7; i++)
  {
    buffer[i] = StartSequence[i * 2 + 1] * uint16_t(256) + StartSequence[i * 2];
    //buffer[i] = 0B1111000000001111;  // 15 240 (sending low byte first)
  }
  BufferWriteIndex = 8;  //  8 means 16 bytes
  BufferReadIndex = 0;
  // -------------------------------------------------- Start ADC
  AdcDummyCounter = 0;
  AdcCounter = 0;
  ADCSRA |= (1<<ADEN);  // Enable ADC
  ADCSRA |= (1<<ADATE); // Enable ADC free-running
  ADCSRA |= (1<<ADSC);  // Start ADC conversion
  // --------------------------------------------------
  interrupts();
}

// ==================================================================
//   SEND SAMPLES
// ==================================================================
void SendSamples()
{
  while (BufferReadIndex != BufferWriteIndex)
  {
    Serial.write(buffer[BufferReadIndex] & 255);
    Serial.write(buffer[BufferReadIndex] >> 8);
    if (++BufferReadIndex >= BUFFER_SIZE) BufferReadIndex = 0;
  }
}

// ============================================================================
//  READ AND DECODE SERIAL COMMANDS
// ============================================================================
String ReceivedString;

void ReadCommands()
{
  while (Serial.available()) 
  {
    char inChar = (char)Serial.read(); // get the new byte:
    ReceivedString += inChar;          // add it to the inputString:
    if (inChar == '\n')                // test for newline
    {
      ReceivedString.trim();
      //Serial.println("RECEIVED STRING: " + ReceivedString);
      ReceivedString.toUpperCase();
      // ------------------------------------------------------------ Decode commands
      if (ReceivedString.startsWith("OPTIONS ")) DecodeOptions(ReceivedString);
      else 
      {
      }
      // ------------------------------------------------------------ Empty ReceivedString
      ReceivedString = "";
    }
  }
}

// ============================================================================
//  DECODE SERIAL COMMANDS
// ============================================================================
const byte NUM_STRINGS = 8; // was 16
const byte NUM_CHARS   = 16; 
char Tokens[NUM_STRINGS][NUM_CHARS];

void Tokenizer(String str)
{
  const byte MAX_CHARS = 64; // was 128
  str.toUpperCase();
  str.replace("  ", " ");
  char buffer[MAX_CHARS];
  str.toCharArray(buffer, MAX_CHARS);
  char* token = strtok(buffer, " ");
  int i = 0;
  while (token != NULL && i < NUM_STRINGS)
  { 
    strncpy(Tokens[i], token, NUM_CHARS - 1);
    token = strtok(NULL, " "); 
    i++;
  }
}

void DecodeOptions(String str)
{
  // ---------------------------------------------- Tokenize
  Tokenizer(str);
  // ---------------------------------------------- Decode NumSamples and AdcSpeed
  uint16_t ns = atoi(Tokens[1]);
  uint8_t  as = atoi(Tokens[2]);
  // ---------------------------------------------- Decode ExposureTime us/ms/sec/min 
  float f = atof(Tokens[3]);
  String s = (String)Tokens[3];
  s.toLowerCase();
  if (s.endsWith("ms"))  f *=     1000;
  if (s.endsWith("sec")) f *=  1000000;
  if (s.endsWith("min")) f *= 60000000;
  ExposureTime = f;
  // ---------------------------------------------- Decode DebugType
  DebugType = atoi(Tokens[4]);
  // ---------------------------------------------- Decode SensorType
  uint8_t st = atoi(Tokens[5]);
  // ---------------------------------------------- Set params if changed
  if (ns != NumSamples || 
      as != AdcSpeed || 
      st != SensorType)
  {
    NumSamples = ns;
    AdcSpeed = as;
    SensorType = st;
    SetNumPixelsAndAdcSpeed();
  } 
  // --------------------------------------------- Decomment to debug 
  //Serial.println(String(ExposureTime));
  //ExposureTime = 999999999;
}
